'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var fs = require('fs-extra');
var path = require('path');
var child_process = require('child_process');
var androidToolsBin = require('android-tools-bin');
var EventEmitter = require('events');
var cp = require('cancelable-promise');

function _interopDefaultLegacy (e) { return e && typeof e === 'object' && 'default' in e ? e : { 'default': e }; }

var fs__default = /*#__PURE__*/_interopDefaultLegacy(fs);
var path__default = /*#__PURE__*/_interopDefaultLegacy(path);
var child_process__default = /*#__PURE__*/_interopDefaultLegacy(child_process);
var EventEmitter__default = /*#__PURE__*/_interopDefaultLegacy(EventEmitter);
var cp__default = /*#__PURE__*/_interopDefaultLegacy(cp);

/*
 * Copyright (C) 2017-2020 UBports Foundation <info@ubports.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * Remove falsy values
 * @param {Object} obj object to process
 */
function removeFalsy(obj) {
  if (typeof obj !== "object" || Array.isArray(obj)) return obj;
  for (var i in obj) {
    if (!obj[i]) {
      delete obj[i];
    } else {
      obj[i] = removeFalsy(obj[i]);
      if (!obj[i]) {
        delete obj[i];
      }
    }
  }
  return Object.keys(obj).length ? obj : null;
}

const CancelablePromise = cp__default['default']?.CancelablePromise
  ? cp__default['default'].CancelablePromise
  : cp__default['default'];

/**
 * generic tool class
 * @property {String} tool tool identifier
 * @property {String} executable tool executable path
 * @property {Array<String>} extra extra cli arguments
 * @property {Object} execOptions options for child_process.exec
 */
class Tool extends EventEmitter__default['default'] {
  constructor(options = {}) {
    super();
    this.tool = options?.tool;
    this.executable = androidToolsBin.getAndroidToolPath(options?.tool);
    this.extra = options?.extra || [];
    this.execOptions = options?.execOptions || {};
    this.processes = [];
  }

  /**
   * Terminate all child processes with extreme prejudice.
   */
  kill() {
    this.processes.forEach(child => child.kill()); // HE EVEN KILLED THE YOUNGLINGS
  }

  /**
   * Execute a command. Used for short operations and operations that do not require real-time data access.
   * @param  {...any} args tool arguments
   * @private
   * @returns {CancelablePromise<String>} stdout
   */
  exec(...args) {
    const _this = this;
    return new CancelablePromise((resolve, reject, onCancel) => {
      const cp = child_process__default['default'].execFile(
        _this.executable,
        [..._this.extra, ...args],
        _this.execOptions,
        (error, stdout, stderr) => {
          _this.emit(
            "exec",
            removeFalsy({
              cmd: [_this.tool, ..._this.extra, ...args],
              error: error
                ? {
                    message: error?.message
                      ?.replace(new RegExp(this.executable, "g"), this.tool)
                      .trim(),
                    code: error?.code,
                    signal: error?.signal,
                    killed: error?.killed
                  }
                : null,
              stdout: stdout?.trim(),
              stderr: stderr?.trim()
            })
          );
          if (error) {
            reject(new Error(_this.handleError(error, stdout, stderr)));
          } else {
            resolve(stdout?.trim() || stderr?.trim());
            this.processes.splice(this.processes.indexOf(cp), 1);
          }
        }
      );

      this.processes.push(cp);

      onCancel(() => {
        if (!cp.kill("SIGTERM")) {
          setTimeout(() => {
            cp.kill("SIGKILL");
          }, 25);
        }
      });
    });
  }

  /**
   * Spawn a child process. Used for long-running operations that require real-time data access.
   * @param  {...any} args tool arguments
   * @private
   * @returns {child_process.ChildProcess}
   */
  spawn(...args) {
    this.emit(
      "spawn:start",
      removeFalsy({ cmd: [this.tool, ...this.extra, ...args].flat() })
    );
    const cp = child_process__default['default'].spawn(
      this.executable,
      [...this.extra, ...args].flat(),
      {
        env: {
          ADB_TRACE: "rwx"
        }
      }
    );
    this.processes.push(cp);
    cp.on("exit", (code, signal) => {
      this.processes.splice(this.processes.indexOf(cp), 1);
      this.emit(
        "spawn:exit",
        removeFalsy({
          cmd: [this.tool, ...this.extra, ...args].flat(),
          code,
          signal
        })
      );
    });
    cp.on("error", error =>
      this.emit(
        "spawn:error",
        removeFalsy({
          cmd: [this.tool, ...this.extra, ...args].flat(),
          error
        })
      )
    );
    return cp;
  }

  /**
   * Generate processable error messages from child_process.exec() callbacks
   * @param {child_process.ExecException} error error returned by child_process.exec()
   * @param {String} stdout stdandard output
   * @param {String} stderr standard error
   * @private
   * @returns {String} error message
   */
  handleError(error, stdout, stderr) {
    if (
      stderr?.includes("Killed") ||
      stderr?.includes("killed by remote request")
    ) {
      return "killed";
    } else {
      return JSON.stringify(
        removeFalsy({
          error: removeFalsy(error),
          stdout: stdout?.trim(),
          stderr: stderr?.trim()
        })
      ).replace(new RegExp(this.executable, "g"), this.tool);
    }
  }

  /**
   * Find out if a device can be seen
   * @virtual
   * @returns {CancelablePromise<Boolean>} access?
   */
  hasAccess() {
    return CancelablePromise.reject(new Error("virtual"));
  }

  /**
   * Wait for a device
   * @returns {CancelablePromise}
   */
  wait() {
    var _this = this;
    return new CancelablePromise(function (resolve, reject, onCancel) {
      let timeout;
      let stop;
      function poll() {
        if (!stop) {
          _this
            .hasAccess()
            .then(access => {
              if (access) {
                clearTimeout(timeout);
                resolve();
              } else {
                timeout = setTimeout(poll, 2000);
              }
            })
            .catch(error => {
              if (error) {
                clearTimeout(timeout);
                reject(error);
              }
            });
        } else {
          clearTimeout(timeout);
        }
      }
      onCancel(() => {
        clearTimeout(timeout);
        stop = true;
      });
      poll();
    });
  }
}

const SERIALNO = /^([0-9]|[a-z])+([0-9a-z]+)$/i;
const DEFAULT_PORT = 5037;

/**
 * Android Debug Bridge (ADB) module
 */
class Adb extends Tool {
  constructor(options) {
    super({
      tool: "adb",
      extra: ["-P", options?.port || DEFAULT_PORT],
      ...options
    });
    this.port = options?.port || DEFAULT_PORT;
  }

  /**
   * Terminate all child processes with extreme prejudice.
   */
  kill() {
    this.killServer().finally(() => super.kill());
  }

  /**
   * Generate processable error messages from child_process.exec() callbacks
   * @param {child_process.ExecException} error error returned by child_process.exec()
   * @param {String} stdout stdandard output
   * @param {String} stderr standard error
   * @private
   * @returns {String} error message
   */
  handleError(error, stdout, stderr) {
    if (
      stderr?.includes("error: device unauthorized") ||
      stderr?.includes("error: device still authorizing")
    ) {
      return "unauthorized";
    } else if (stderr?.includes("error: device offline")) {
      return "device offline";
    } else if (
      stderr?.includes("no devices/emulators found") ||
      stdout?.includes("no devices/emulators found") ||
      stdout?.includes("adb: error: failed to read copy response") ||
      stdout?.includes("couldn't read from device") ||
      stdout?.includes("remote Bad file number") ||
      stdout?.includes("remote Broken pipe") ||
      stderr?.includes("adb: sideload connection failed: closed") ||
      stderr?.includes("adb: pre-KitKat sideload connection failed: closed")
    ) {
      return "no device";
    } else {
      return super.handleError(error, stdout, stderr);
    }
  }

  /**
   * Kill all adb servers and start a new one to rule them all
   * @returns {Promise}
   */
  startServer() {
    return this.killServer().then(() => this.exec("start-server"));
  }

  /**
   * Kill all running servers
   * @returns {Promise}
   */
  killServer() {
    return this.exec("kill-server");
  }

  /**
   * kick connection from host side to force reconnect
   * @param {String} [modifier] - "device" or "offline"
   * @returns {Promise<String>} resolves device state
   */
  reconnect(modifier = "") {
    return this.exec(...["reconnect", modifier].filter(i => i)).then(stdout => {
      if (stdout?.includes("no devices/emulators found")) {
        throw new Error("no device");
      } else {
        return this.wait();
      }
    });
  }

  /**
   * kick connection from device side to force reconnect
   * @returns {Promise<String>} resolves device state
   */
  reconnectDevice() {
    return this.reconnect("device");
  }

  /**
   * reset offline/unauthorized devices to force reconnect
   * @returns {Promise<String>} resolves device state
   */
  reconnectOffline() {
    return this.reconnect("offline");
  }

  /**
   * Get the devices serial number
   * @returns {Promise<String>} serial number
   */
  getSerialno() {
    return this.exec("get-serialno").then(stdout => {
      if (!stdout || stdout?.includes("unknown") || !SERIALNO.test(stdout)) {
        throw new Error(`invalid serial number: ${stdout?.trim()}`);
      } else {
        return stdout.trim();
      }
    });
  }

  /**
   * run remote shell command
   * @param {Array} args - list of shell arguments
   * @returns {Promise<String>} stdout
   */
  shell(...args) {
    return this.exec("shell", args.join(" ")).then(stdout => stdout?.trim());
  }

  /**
   * copy local files/directories to device
   * @param {Array<String>} files path to files
   * @param {String} dest destination path on the device
   * @param {Function} progress progress function
   * @returns {CancelablePromise}
   */
  push(files = [], dest, progress = () => {}) {
    progress(0);
    if (!files?.length) {
      // if there are no files, report 100% and resolve
      progress(1);
      return Promise.resolve();
    } else {
      const totalSize = files.reduce(
        (acc, file) => acc + fs__default['default'].statSync(file)["size"],
        0
      );
      let pushedSize = 0;
      const _this = this;
      return new CancelablePromise((resolve, reject, onCancel) => {
        let stdout = "";
        let stderr = "";
        const cp = _this.spawn("push", ...files, dest);
        cp.once("exit", (code, signal) => {
          if (code || signal) {
            // truthy value (i.e. non-zero exit code) indicates error
            if (
              stdout.includes("adb: error: cannot stat") &&
              stdout.includes("No such file or directory")
            ) {
              reject(new Error("file not found"));
            } else {
              reject(
                new Error(_this.handleError({ code, signal }, stdout, stderr))
              );
            }
          } else {
            resolve();
          }
        });

        cp.stdout.on("data", d => (stdout += d.toString()));
        cp.stderr.on("data", d => {
          d.toString()
            .split("\n")
            .forEach(str => {
              if (str.includes("cpp")) {
                // logging from cpp files indicates debug output
                if (str.includes("writex")) {
                  // writex namespace indicates external writing
                  pushedSize +=
                    parseInt(str?.split("len=")[1]?.split(" ")[0]) || 0;
                  progress(Math.min(pushedSize / totalSize, 1));
                }
              } else {
                stderr += str;
              }
            });
        });

        onCancel(() => {
          if (!cp.kill("SIGTERM")) {
            setTimeout(() => cp.kill("SIGKILL"), 25);
          }
        });
      });
    }
  }

  /**
   * Reboot to a state
   * @param {String} state - system, recovery, bootloader
   * @async
   * @returns {Promise}
   */
  async reboot(state) {
    if (!["system", "recovery", "bootloader"].includes(state)) {
      throw new Error("unknown state: " + state);
    } else {
      return this.exec("reboot", state).then(stdout => {
        if (stdout?.includes("failed"))
          throw new Error(`reboot failed: ${stdout}`);
        else return;
      });
    }
  }

  /**
   * sideload an ota package
   * @param {String} file - path to a file to sideload
   * @returns {CancelablePromise}
   */
  sideload(file, progress = () => {}) {
    progress(0);
    const totalSize = fs__default['default'].statSync(file)["size"];
    let pushedSize = 0;
    const _this = this;
    return new CancelablePromise((resolve, reject, onCancel) => {
      let stdout = "";
      let stderr = "";
      const cp = _this.spawn("sideload", file);
      cp.once("exit", (code, signal) => {
        if (code || signal) {
          // truthy value (i.e. non-zero exit code) indicates error
          if (
            stdout.includes("adb: error: cannot stat") &&
            stdout.includes("No such file or directory")
          ) {
            reject(new Error("file not found"));
          } else {
            reject(
              new Error(_this.handleError({ code, signal }, stdout, stderr))
            );
          }
        } else {
          resolve();
        }
      });

      cp.stdout.on("data", d => (stdout += d.toString()));
      cp.stderr.on("data", d => {
        d.toString()
          .split("\n")
          .filter(s => s)
          .forEach(str => {
            if (str.includes("cpp")) {
              // logging from cpp files indicates debug output
              if (str.includes("writex")) {
                // writex namespace indicates external writing
                pushedSize += parseInt(str.split("len=")[1].split(" ")[0]) || 0;
                progress(Math.min(pushedSize / totalSize, 1));
              }
            } else {
              stderr += str;
            }
          });
      });

      onCancel(() => {
        if (!cp.kill("SIGTERM")) {
          setTimeout(() => cp.kill("SIGKILL"), 25);
        }
      });
    });
  }

  /**
   * Return the status of the device
   * @returns {Promise<String>} bootloader, recovery, device
   */
  getState() {
    return this.exec("get-state").then(stdout => stdout.trim());
  }

  //////////////////////////////////////////////////////////////////////////////
  // Convenience functions
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Reboot to a requested state, if not already in it
   * @param {String} state - system, recovery, bootloader
   * @returns {Promise}
   */
  ensureState(state) {
    return this.getState().then(currentState =>
      currentState === state ||
      (currentState === "device" && state === "system")
        ? Promise.resolve()
        : this.reboot(state).then(() => this.wait())
    );
  }

  /**
   * get device codename from getprop or by reading the default.prop file
   * @returns {Promise<String>} codename
   */
  getDeviceName() {
    return this.shell("getprop", "ro.product.device").then(stdout => {
      if (!stdout || stdout?.includes("getprop: not found")) {
        return this.shell("cat", "default.prop")
          .catch(e => {
            throw new Error("getprop error: " + e);
          })
          .then(stdout => {
            if (stdout && stdout.includes("ro.product.device=")) {
              return stdout
                .split("ro.product.device=")[1]
                .split("\n")[0]
                .trim();
            } else {
              throw new Error("unknown getprop error");
            }
          });
      } else {
        return stdout.trim();
      }
    });
  }

  /**
   * Find out what operating system the device is running (currently android and ubuntu touch)
   * @returns {Promise<String>} ubuntutouch, android
   */
  getOs() {
    return this.shell("cat", "/etc/system-image/channel.ini").then(stdout => {
      return stdout ? "ubuntutouch" : "android";
    });
  }

  /**
   * Find out if a device can be seen by adb
   * @returns {Promise<Boolean>} access?
   */
  hasAccess() {
    return this.shell("echo", ".")
      .then(stdout => {
        if (stdout == ".") return true;
        else throw new Error("unexpected response: " + stdout);
      })
      .catch(error => {
        if (error.message && error.message.includes("no device")) {
          return false;
        } else {
          throw error;
        }
      });
  }

  /**
   * wait for device to be in a given state
   * @param {String} [state] any, device, recovery, rescue, sideload, bootloader, or disconnect
   * @param {String} [transport] any, usb, local
   * @returns {CancelablePromise<String>} resolves state (offline, bootloader, device)
   */
  wait(state = "any", transport = "any") {
    if (
      ![
        "any",
        "device",
        "recovery",
        "rescue",
        "sideload",
        "bootloader",
        "disconnect"
      ].includes(state)
    ) {
      return Promise.reject(new Error(`Invalid state: ${state}`));
    } else if (!["any", "usb", "local"].includes(transport)) {
      return Promise.reject(new Error(`Invalid transport: ${transport}`));
    }
    return this.exec(`wait-for-${transport}-${state}`).then(() =>
      this.getState()
    );
  }

  /**
   * Format partition
   * @param {String} partition partition to format
   * @returns {Promise}
   */
  format(partition) {
    return this.shell("cat", "/etc/recovery.fstab")
      .then(fstab => {
        if (!fstab || typeof fstab !== "string") {
          throw new Error("unable to read recovery.fstab");
        } else {
          const block = this.findPartitionInFstab(partition, fstab);
          return this.shell("umount /" + partition)
            .then(() => this.shell("make_ext4fs " + block))
            .then(() => this.shell("mount /" + partition))
            .then(error => {
              if (error) throw new Error("failed to mount: " + error);
              else return;
            });
        }
      })
      .catch(error => {
        throw new Error("failed to format " + partition + ": " + error);
      });
  }

  /**
   * If cache can not be formated, rm it
   * @returns {Promise}
   */
  wipeCache() {
    return this.format("cache").finally(() =>
      this.shell("rm", "-rf", "/cache/*")
    );
  }

  /**
   * Find the partition associated with a mountpoint in an fstab
   * @param {String} partition partition to find
   * @param {String} fstab fstab
   * @returns {String} partition
   */
  findPartitionInFstab(partition, fstab) {
    try {
      return fstab
        .split("\n")
        .filter(block => block.startsWith("/dev"))
        .filter(
          block => block.split(" ").filter(c => c !== "")[1] === "/" + partition
        )[0]
        .split(" ")[0];
    } catch (error) {
      throw new Error("failed to parse fstab");
    }
  }

  /**
   * Find a partition and verify its type
   * @param {String} partition partition to verify
   * @param {String} type expected type
   * @returns {Promise<Boolean>} verified?
   */
  verifyPartitionType(partition, type) {
    return this.shell("mount").then(stdout => {
      if (
        !(stdout.includes(" on /") && stdout.includes(" type ")) ||
        typeof stdout !== "string" ||
        !stdout.includes("/" + partition)
      ) {
        throw new Error("partition not found");
      } else {
        return stdout.includes(" on /" + partition + " type " + type);
      }
    });
  }

  /**
   * size of a file or directory
   * @param {String} file file or directory
   * @returns {Promise<Float>} size
   */
  getFileSize(file) {
    return this.shell("du -shk " + file)
      .then(size => {
        if (isNaN(parseFloat(size)))
          throw new Error(`Cannot parse size from ${size}`);
        else return parseFloat(size);
      })
      .catch(e => {
        throw new Error(`Unable to get size: ${e}`);
      });
  }

  /**
   * available size of a partition
   * @param {String} partition partition to check
   * @returns {Promise<Integer>} available size
   */
  getAvailablePartitionSize(partition) {
    return this.shell("df -k -P " + partition)
      .then(stdout => stdout.split(/[ ,]+/))
      .then(arr => parseInt(arr[arr.length - 3]))
      .then(size => {
        if (isNaN(size)) throw new Error(`Cannot parse size from ${size}`);
        else return size;
      })
      .catch(e => {
        throw new Error(`Unable to get size: ${e}`);
      });
  }

  /**
   * total size of a partition
   * @param {String} partition partition to check
   * @returns {Promise<Integer>} total size
   */
  getTotalPartitionSize(partition) {
    return this.shell("df -k -P " + partition)
      .then(stdout => stdout.split(/[ ,]+/))
      .then(arr => parseInt(arr[arr.length - 5]))
      .then(size => {
        if (isNaN(size)) throw new Error(`Cannot parse size from ${size}`);
        else return size;
      })
      .catch(e => {
        throw new Error(`Unable to get size: ${e}`);
      });
  }

  /**
   * [EXPERIMENTAL] Run a command via adb exec-out and pipe the result to a stream
   * @param {steam.writable} writableStream e.g. fs.createWriteStream() to write to a file
   * @param  {...any} args command to execute
   * @returns {Promise}
   */
  execOut(writableStream, ...args) {
    const _this = this;
    return new Promise(function (resolve, reject) {
      let stderr = "";
      const cp = _this.spawn("exec-out", `'${args.join(" ")}'`);
      cp.stdout.pipe(writableStream);
      cp.stderr.on("data", d => (stderr += d.toString()));
      cp.once("exit", (code, signal) => {
        writableStream.close();
        if (code || signal)
          reject(new Error(_this.handleError({ code, signal }, null, stderr)));
        else resolve();
      });
    });
  }

  /**
   * [EXPERIMENTAL] Backup "srcfile" from the device to local tar "destfile"
   * @param {String} srcfile file to back up
   * @param {String} destfile target destination
   * @param {Function} progress progress function
   * @async
   * @returns {Promise}
   */
  async createBackupTar(srcfile, destfile, progress) {
    progress(0);
    const stream = fs__default['default'].createWriteStream(destfile);
    const fileSize = await this.getFileSize(srcfile);
    // FIXME with gzip compression (the -z flag on tar), the progress estimate is way off. It's still beneficial to enable it, because it saves a lot of space.
    let timeout;
    function poll() {
      const { size } = fs__default['default'].statSync(destfile);
      progress(size / 1024 / fileSize);
      timeout = setTimeout(poll, 1000);
    }
    poll();
    return this.execOut(
      stream,
      "tar",
      "-cpz",
      "--exclude=*/var/cache ",
      "--exclude=*/var/log ",
      "--exclude=*/.cache/upstart ",
      "--exclude=*/.cache/*.qmlc ",
      "--exclude=*/.cache/*/qmlcache ",
      "--exclude=*/.cache/*/qml_cache",
      srcfile,
      "2>/dev/null"
    ).finally(() => clearTimeout(timeout));
  }

  /**
   * [EXPERIMENTAL] Restore tar "srcfile"
   * @param {String} srcfile file to restore
   * @param {Function} progress progress callback
   * @returns {Promise}
   */
  restoreBackupTar(srcfile, progress = () => {}) {
    progress();
    return this.ensureState("recovery")
      .then(() => this.shell("mkfifo /restore.pipe"))
      .then(() =>
        Promise.all([
          this.push([srcfile], "/restore.pipe", progress),
          this.shell("cd /;", "cat /restore.pipe | tar -xvz")
        ])
      )
      .then(() => this.shell("rm", "/restore.pipe"))
      .catch(e => {
        throw new Error(`Restore failed: ${e}`);
      });
  }

  /**
   * List backups
   * @param {String} backupBaseDir path to backup storage location
   * @returns {Promise<Array<Object>>} backup list
   */
  listUbuntuBackups(backupBaseDir) {
    return fs__default['default']
      .readdir(backupBaseDir)
      .then(backups =>
        Promise.all(
          backups.map(backup =>
            fs__default['default']
              .readJSON(path__default['default'].join(backupBaseDir, backup, "metadata.json"))
              .then(metadata => ({
                ...metadata,
                dir: path__default['default'].join(backupBaseDir, backup)
              }))
              .catch(() => null)
          )
        ).then(r => r.filter(r => r))
      )
      .catch(() => []);
  }

  /**
   * create a full backup of ubuntu touch
   * @param {String} backupBaseDir path to backup storage location
   * @param {String} [comment] description of the backup
   * @param {String} [dataPartition="/data"] data partition on the device
   * @param {Function} [progress] progress function
   * @returns {Promise<Object>} backup object
   */
  async createUbuntuTouchBackup(
    backupBaseDir,
    comment,
    dataPartition = "/data",
    progress = () => {}
  ) {
    progress(0);
    const time = new Date();
    const dir = path__default['default'].join(backupBaseDir, time.toISOString());
    return this.ensureState("recovery")
      .then(() => fs__default['default'].ensureDir(dir))
      .then(() =>
        Promise.all([
          this.shell("stat", "/data/user-data"),
          this.shell("stat", "/data/syste-mdata")
        ]).catch(() => this.shell("mount", dataPartition, "/data"))
      )
      .then(() =>
        this.createBackupTar(
          "/data/system-data",
          path__default['default'].join(dir, "system.tar.gz"),
          p => progress(p * 0.5)
        )
      )
      .then(() =>
        this.createBackupTar(
          "/data/user-data",
          path__default['default'].join(dir, "user.tar.gz"),
          p => progress(50 + p * 0.5)
        )
      )
      .then(async () => {
        const metadata = {
          codename: await this.getDeviceName(),
          serialno: await this.getSerialno(),
          size:
            (await this.getFileSize("/data/user-data")) +
            (await this.getFileSize("/data/system-data")),
          time,
          comment:
            comment || `Ubuntu Touch backup created on ${time.toISOString()}`,
          restorations: []
        };
        return fs__default['default']
          .writeJSON(path__default['default'].join(dir, "metadata.json"), metadata)
          .then(() => ({ ...metadata, dir }))
          .catch(e => {
            throw new Error(`Failed to restore: ${e}`);
          });
      });
  }

  /**
   * restore a full backup of ubuntu touch
   * @param {String} dir directory where backup is stored
   * @param {Function} [progress] progress function
   * @returns {Promise<Object>} backup object
   */
  async restoreUbuntuTouchBackup(dir, progress = () => {}) {
    progress(0); // FIXME report actual push progress
    let metadata = await fs__default['default'].readJSON(path__default['default'].join(dir, "metadata.json"));
    return this.ensureState("recovery")
      .then(async () => {
        metadata.restorations = metadata.restorations || [];
        metadata.restorations.push({
          codename: await this.getDeviceName(),
          serialno: await this.getSerialno(),
          time: new Date().toISOString()
        });
      })
      .then(() => progress(10))
      .then(() => this.restoreBackupTar(path__default['default'].join(dir, "system.tar.gz")))
      .then(() => progress(50))
      .then(() => this.restoreBackupTar(path__default['default'].join(dir, "user.tar.gz")))
      .then(() => progress(90))
      .then(() => fs__default['default'].writeJSON(path__default['default'].join(dir, "metadata.json"), metadata))
      .then(() => this.reboot("system"))
      .then(() => progress(100))
      .then(() => ({ ...metadata, dir }))
      .catch(e => {
        throw new Error(`Failed to restore: ${e}`);
      });
  }
}

/**
 * fastboot android flashing and booting utility
 */
class Fastboot extends Tool {
  constructor(options) {
    super({
      tool: "fastboot",
      ...options
    });
  }

  /**
   * Generate processable error messages from child_process.exec() callbacks
   * @param {child_process.ExecException} error error returned by child_process.exec()
   * @param {String} stdout stdandard output
   * @param {String} stderr standard error
   * @private
   * @returns {String} error message
   */
  handleError(error, stdout, stderr) {
    if (
      stderr?.includes("FAILED (remote: low power, need battery charging.)")
    ) {
      return "low battery";
    } else if (
      stderr?.includes("not supported in locked device") ||
      stderr?.includes("Bootloader is locked") ||
      stderr?.includes("not allowed in locked state") ||
      stderr?.includes("Device not unlocked cannot flash or erase") ||
      stderr?.includes("Partition flashing is not allowed") ||
      stderr?.includes("Command not allowed") ||
      stderr?.includes("not allowed when locked") ||
      stderr?.includes("device is locked. Cannot flash images") ||
      stderr?.match(/download for partition '[a-z]+' is not allowed/i)
    ) {
      return "bootloader locked";
    } else if (
      stderr?.includes("Check 'Allow OEM Unlock' in Developer Options") ||
      stderr?.includes("Unlock operation is not allowed") ||
      stderr?.includes("oem unlock is not allowed")
    ) {
      return "enable unlocking";
    } else if (stderr?.includes("FAILED (remote failure)")) {
      return "failed to boot";
    } else if (
      stderr?.includes("I/O error") ||
      stderr?.includes("FAILED (command write failed (No such device))") ||
      stderr?.includes("FAILED (command write failed (Success))") ||
      stderr?.includes("FAILED (status read failed (No such device))") ||
      stderr?.includes("FAILED (data transfer failure (Broken pipe))") ||
      stderr?.includes("FAILED (data transfer failure (Protocol error))")
    ) {
      return "no device";
    } else {
      return super.handleError(error, stdout, stderr);
    }
  }

  /**
   * @typedef FastbootFlashImage
   * @property {String} partition partition to flash
   * @property {String} file path to an image file
   * @property {Boolean} raw use `fastboot flash:raw` instead of `fastboot flash`
   * @property {Array<String>} flags additional cli-flags like --force and --disable-verification
   */

  /**
   * Write a file to a flash partition
   * @param {Array<FastbootFlashImage>} images Images to flash
   * @param {Function} progress progress callback
   * @returns {Promise}
   */
  flash(images, progress = () => {}) {
    progress(0);
    const _this = this;
    // build a promise chain to flash all images sequentially
    return images
      .reduce(
        (prev, image, i) =>
          prev.then(
            () =>
              new Promise((resolve, reject) => {
                let stdout = "";
                let stderr = "";
                let offset = i / images.length;
                let scale = 1 / images.length;
                let sparseCurr = 1;
                let sparseTotal = 1;
                const cp = _this.spawn(
                  image.raw ? "flash:raw" : "flash",
                  image.partition,
                  ...(image?.flags || []),
                  image.file
                );
                cp.once("exit", (code, signal) => {
                  if (code || signal) {
                    reject(_this.handleError({ code, signal }, stdout, stderr));
                  } else {
                    resolve();
                  }
                });
                cp.stdout.on("data", d => (stdout += d.toString()));
                cp.stderr.on("data", d => {
                  d.toString()
                    .trim()
                    .split("\n")
                    .forEach(str => {
                      // FIXME improve and simplify logic
                      if (!str.includes("OKAY")) {
                        if (str.includes("Sending")) {
                          try {
                            if (str.includes("sparse")) {
                              sparseCurr = parseInt(
                                str.split("/")[0].split("' ")[1]
                              );
                              sparseTotal = parseInt(
                                str.split("/")[1].split(" ")[0]
                              );
                            }
                          } catch {}
                          progress(
                            offset + scale * ((sparseCurr * 0.3) / sparseTotal)
                          );
                        } else if (str.includes("Writing")) {
                          progress(
                            offset + scale * ((sparseCurr * 0.9) / sparseTotal)
                          );
                        } else if (!str.includes("Finished")) {
                          stderr += str;
                        }
                      }
                    });
                });
              })
          ),
        _this.wait()
      )
      .then(() => progress(1))
      .catch(e => {
        throw new Error(`Flashing failed: ${e}`);
      });
  }

  /**
   * Download and boot kernel
   * @param {String} image image file to boot
   * @returns {Promise}
   */
  boot(image) {
    return this.exec("boot", image)
      .then(stdout => {
        return;
      })
      .catch(error => {
        throw new Error("booting failed: " + error);
      });
  }

  /**
   * Reflash device from update.zip and set the flashed slot as active
   * @param {String} image image to flash
   * @param {String} wipe wipe option
   * @returns {Promise}
   */
  update(image, wipe) {
    return this.exec(wipe ? "-w" : "", "update", image)
      .then(stdout => {
        return;
      })
      .catch(error => {
        throw new Error("update failed: " + error);
      });
  }

  /**
   * Reboot device into bootloader
   * @returns {Promise}
   */
  rebootBootloader() {
    return this.exec("reboot-bootloader")
      .then(() => {
        return;
      })
      .catch(error => {
        throw new Error("rebooting to bootloader failed: " + error);
      });
  }

  /**
   * Reboot device
   * @returns {Promise}
   */
  reboot() {
    return this.exec("reboot")
      .then(() => {
        return;
      })
      .catch(error => {
        throw new Error("rebooting failed: " + error);
      });
  }

  /**
   * Continue with autoboot
   * @returns {Promise}
   */
  continue() {
    return this.exec("continue")
      .then(() => {
        return;
      })
      .catch(error => {
        throw new Error("continuing boot failed: " + error);
      });
  }

  /**
   * Format a flash partition. Can override the fs type and/or size the bootloader reports.
   * @param {String} partition to format
   * @param {String} type partition type
   * @param {String} size partition size
   * @returns {Promise}
   */
  format(partition, type, size) {
    if (!type && size) {
      return Promise.reject(
        new Error(
          "formatting failed: size specification requires type to be specified as well"
        )
      );
    }
    return this.exec(
      `format${type ? ":" + type : ""}${size ? ":" + size : ""}`,
      partition
    )
      .then(() => {
        return;
      })
      .catch(error => {
        throw new Error("formatting failed: " + error);
      });
  }

  /**
   * Erase a flash partition
   * @param {String} partition partition to erase
   * @returns {Promise}
   */
  erase(partition) {
    return this.exec("erase", partition)
      .then(() => {
        return;
      })
      .catch(error => {
        throw new Error("erasing failed: " + error);
      });
  }

  /**
   * Sets the active slot
   * @param {String} slot - slot to set as active
   */
  setActive(slot) {
    return this.exec(`--set-active=${slot}`)
      .then(stdout => {
        if (stdout && stdout.includes("error")) {
          throw new Error(stdout);
        } else {
          return;
        }
      })
      .catch(error => {
        throw new Error(`failed to set active slot: ${error}`);
      });
  }

  //////////////////////////////////////////////////////////////////////////////
  // Convenience functions
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Lift OEM lock
   * @param {String} [code] optional unlock code (including 0x if necessary)
   * @returns {Promise}
   */
  oemUnlock(code = "") {
    return this.exec(...["oem", "unlock", code || []].flat())
      .then(() => null)
      .catch(error => {
        if (
          error?.message.includes("Already Unlocked") ||
          error?.message.includes("Not necessary")
        )
          return;
        else throw new Error("oem unlock failed: " + error);
      });
  }

  /**
   * Enforce OEM lock
   * @returns {Promise}
   */
  oemLock() {
    return this.exec("oem", "lock")
      .then(() => null)
      .catch(error => {
        throw new Error("oem lock failed: " + error);
      });
  }

  /**
   * lock partitions for flashing
   * @returns {Promise}
   */
  flashingLock() {
    return this.exec("flashing", "lock").then(() => null);
  }

  /**
   * unlock partitions for flashing
   * @returns {Promise}
   */
  flashingUnlock() {
    return this.exec("flashing", "unlock").then(() => null);
  }

  /**
   * lock 'critical' bootloader partitions
   * @returns {Promise}
   */
  flashingLockCritical() {
    return this.exec("flashing", "lock_critical").then(() => null);
  }

  /**
   * unlock 'critical' bootloader partitions
   * @returns {Promise}
   */
  flashingUnlockCritical() {
    return this.exec("flashing", "unlock_critical").then(() => null);
  }

  /**
   * Find out if a device can be flashing-unlocked
   * @returns {Promise<Boolean>}
   */
  getUnlockAbility() {
    return this.exec("flashing", "get_unlock_ability")
      .then(stdout => stdout?.trim() === "1")
      .catch(() => false);
  }

  /**
   * Find out if a device can be seen by fastboot
   * @returns {Promise}
   */
  hasAccess() {
    return this.exec("devices")
      .then(stdout => {
        return Boolean(stdout && stdout.includes("fastboot"));
      })
      .catch(error => {
        throw error;
      });
  }

  /**
   * Wait for a device
   * @returns {CancelablePromise<String>}
   */
  wait() {
    return super.wait().then(() => "bootloader");
  }

  /**
   * get bootloader var
   * @param {String} variable variable to get
   * @returns {Promise<String>} codename
   */
  getvar(variable) {
    return this.hasAccess()
      .then(access => {
        if (access) {
          return this.exec("getvar", variable);
        } else {
          throw new Error("no device");
        }
      })
      .then(r => r.split("\n")[0].split(": "))
      .then(([name, value]) => {
        if (name !== variable) {
          throw new Error(`Unexpected getvar return: ${name}`);
        } else {
          return value;
        }
      });
  }

  /**
   * get device codename from product bootloader var
   * @returns {Promise<String>} codename
   */
  getDeviceName() {
    return this.getvar("product");
  }
}

/**
 * heimdall: flash firmware on samsung devices
 */
class Heimdall extends Tool {
  constructor(options) {
    super({
      tool: "heimdall",
      ...options
    });
  }

  /**
   * Generate processable error messages from child_process.exec() callbacks
   * @param {child_process.ExecException} error error returned by child_process.exec()
   * @param {String} stdout stdandard output
   * @param {String} stderr standard error
   * @private
   * @returns {String} error message
   */
  handleError(error, stdout, stderr) {
    if (
      stderr?.includes(
        "ERROR: Failed to detect compatible download-mode device."
      )
    ) {
      return "no device";
    } else {
      return super.handleError(error, stdout, stderr);
    }
  }

  /**
   * Find out if a device in download mode can be seen by heimdall
   * @returns {Promise<Boolean>}
   */
  detect() {
    return this.hasAccess();
  }

  /**
   * Find out if a device in download mode can be seen by heimdall
   * @returns {Promise<Boolean>}
   */
  hasAccess() {
    return this.exec("detect")
      .then(() => true)
      .catch(error => {
        if (error.message.includes("no device")) {
          return false;
        } else {
          throw error;
        }
      });
  }

  /**
   * Wait for a device
   * @returns {CancelablePromise<String>}
   */
  wait() {
    return super.wait().then(() => "download");
  }

  /**
   * Prints the contents of a PIT file in a human readable format. If a filename is not provided then Heimdall retrieves the PIT file from the connected device.
   * @param {String} file pit file to print
   * @returns {Promise<String>}
   */
  printPit(file) {
    return this.exec("print-pit", ...(file ? ["--file", file] : []))
      .then(r =>
        r
          .split("\n\nEnding session...")[0]
          .split(/--- Entry #\d ---/)
          .slice(1)
          .map(r => r.trim())
      )
      .catch(error => {
        throw error;
      });
  }

  /**
   * get partitions from pit file
   * @returns {Promise<String>}
   */
  getPartitions() {
    return this.printPit().then(r =>
      r.map(r =>
        r
          .split("\n")
          .map(r => r.split(":").map(r => r.trim()))
          .reduce((result, item) => {
            result[item[0]] = item[1];
            return result;
          }, {})
      )
    );
  }

  /**
   * @typedef HeimdallFlashImage
   * @property {String} partition partition to flash
   * @property {String} file path to an image file
   */

  /**
   * Flash firmware files to partitions (names or identifiers)
   * @param {Array<HeimdallFlashImage>} images Images to flash
   * @returns {Promise}
   */
  flash(images) {
    // TODO report progress similar to fastboot.flash()
    return this.exec(
      "flash",
      ...images.map(i => [`--${i.partition}`, i.file]).flat()
    ).then(() => null);
  }
}

/**
 * A wrapper for Adb, Fastboot, and Heimall that returns convenient promises.
 */
class DeviceTools extends EventEmitter__default['default'] {
  constructor() {
    super();
    this.adb = new Adb();
    this.fastboot = new Fastboot();
    this.heimdall = new Heimdall();

    ["adb", "fastboot", "heimdall"].forEach(tool => {
      this[tool].on("exec", r => this.emit("exec", r));
      this[tool].on("spawn:start", r => this.emit("spawn:start", r));
      this[tool].on("spawn:exit", r => this.emit("spawn:exit", r));
      this[tool].on("spawn:error", r => this.emit("spawn:error", r));
    });
  }

  /**
   * Terminate all child processes with extreme prejudice.
   */
  kill() {
    ["adb", "fastboot", "heimdall"].forEach(tool => {
      this[tool].kill();
    });
  }

  /**
   * Wait for a device
   * @returns {CancelablePromise<String>}
   */
  wait() {
    const _this = this;
    return new CancelablePromise(function (resolve, reject, onCancel) {
      const waitPromises = [
        _this.adb.wait(),
        _this.fastboot.wait(),
        _this.heimdall.wait()
      ];
      CancelablePromise.race(waitPromises)
        .then(state => {
          waitPromises.forEach(p => p.cancel());
          resolve(state);
        })
        .catch(() => {
          reject(new Error("no device"));
        });

      onCancel(() => waitPromises.forEach(p => p.cancel()));
    });
  }

  /**
   * Resolve device name
   * @returns {Promise<String>}
   */
  getDeviceName() {
    // TODO support heimdall
    return this.adb
      .getDeviceName()
      .catch(() => this.fastboot.getDeviceName())
      .catch(() => {
        throw new Error("no device");
      });
  }
}

exports.Adb = Adb;
exports.DeviceTools = DeviceTools;
exports.Fastboot = Fastboot;
exports.Heimdall = Heimdall;
exports.Tool = Tool;
